/*
             LUFA Library
     Copyright (C) Dean Camera, 2011.

  dean [at] fourwalledcubicle [dot] com
           www.lufa-lib.org
*/

/*
  Copyright 2011  Dean Camera (dean [at] fourwalledcubicle [dot] com)

  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaim all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/** \file
 *
 *  Main source file for the RNDISEthernet demo. This file contains the main tasks of
 *  the demo and is responsible for the initial application hardware configuration.
 */

#include "RNDISEthernet.h"

//#include <avr/pgmspace.h>
#include <alloca.h>
#include "rfm22b_regs.h"

#define MAX_PACKET_LENGTH 32

/** LUFA RNDIS Class driver interface configuration and state information. This structure is
 *  passed to all RNDIS Class driver functions, so that multiple instances of the same class
 *  within a device can be differentiated from one another.
 */
USB_ClassInfo_RNDIS_Device_t Ethernet_RNDIS_Interface =
	{
		.Config =
			{
				.ControlInterfaceNumber         = 0,

				.DataINEndpointNumber           = CDC_TX_EPNUM,
				.DataINEndpointSize             = CDC_TXRX_EPSIZE,
				.DataINEndpointDoubleBank       = false,

				.DataOUTEndpointNumber          = CDC_RX_EPNUM,
				.DataOUTEndpointSize            = CDC_TXRX_EPSIZE,
				.DataOUTEndpointDoubleBank      = false,

				.NotificationEndpointNumber     = CDC_NOTIFICATION_EPNUM,
				.NotificationEndpointSize       = CDC_NOTIFICATION_EPSIZE,
				.NotificationEndpointDoubleBank = false,

				.AdapterVendorDescription       = "LUFA RNDIS Demo Adapter",
				.AdapterMACAddress              = {ADAPTER_MAC_ADDRESS},
			},
	};

/** LUFA CDC Class driver interface configuration and state information. This structure is
 *  passed to all CDC Class driver functions, so that multiple instances of the same class
 *  within a device can be differentiated from one another. This is for the second CDC interface,
 *  which echos back all received data from the host.
 */
USB_ClassInfo_CDC_Device_t VirtualSerial2_CDC_Interface =
	{
		.Config =
			{
				.ControlInterfaceNumber           = 2,

				.DataINEndpointNumber             = CDC2_TX_EPNUM,
				.DataINEndpointSize               = CDC2_TXRX_EPSIZE,
				.DataINEndpointDoubleBank         = false,

				.DataOUTEndpointNumber            = CDC2_RX_EPNUM,
				.DataOUTEndpointSize              = CDC2_TXRX_EPSIZE,
				.DataOUTEndpointDoubleBank        = false,

				.NotificationEndpointNumber       = CDC2_NOTIFICATION_EPNUM,
				.NotificationEndpointSize         = CDC_NOTIFICATION_EPSIZE,
				.NotificationEndpointDoubleBank   = false,
			},
	};


/** Global to store the incoming frame from the host before it is processed by the device. */
static Ethernet_Frame_Info_t FrameIN;

/** Global to store the outgoing frame created in the device before it is sent to the host. */
static Ethernet_Frame_Info_t FrameOUT;

///////
static FILE USBSerialStream;

void console_sendbyte(char byte);
void console_processbyte(char c);
void console_processline(char *line);
uint8_t parse_ascii(char *word, uint8_t *store, uint8_t max);

int8_t rfm22b_init(void);
uint8_t spi_xfer(uint8_t send);
void rfm22b_writem(uint8_t reg, uint8_t *data, uint8_t count);
void rfm22b_readm(uint8_t reg, uint8_t *data, uint8_t count);
void rfm22b_write(uint8_t reg, uint8_t data);
uint8_t rfm22b_read(uint8_t reg);
void console_break(void);
void rfm22b_reset(void);
void rfm22b_interrupt(void);
void rfm22b_cbi(uint8_t reg, uint8_t bv);
void rfm22b_sbi(uint8_t reg, uint8_t bv);
void rfm22b_cb(uint8_t reg, uint8_t mask);
void rfm22b_sb(uint8_t reg, uint8_t mask);
void tick_1khz(void);
void tick_1hz(void);

static uint8_t noisy=5;
static uint16_t count_1khz=0;
static uint16_t count_for_1hz=0;
static uint16_t rssi_decimate;
static uint8_t rssi_decimate_count;
static uint16_t rssi;

typedef struct {
	unsigned tx_need_ack :1;
	unsigned transmitter_active :1;
	
	uint8_t rx_have_seq;
	uint8_t tx_seq_sent;
	

} radio_status;
typedef struct {
	uint8_t destination;
	uint8_t source;
	uint8_t _type;
	uint8_t sequence;
	uint8_t length;
	uint8_t *data;
} rf_packet;

static radio_status rfm_status;

/** Main program entry point. This routine contains the overall program flow, including initial
 *  setup of all components and the main program loop.
 */
int main(void)
{
	SetupHardware();

	TCP_Init();
	Webserver_Init();

	LEDs_SetAllLEDs(LEDMASK_USB_NOTREADY);
	sei();

	for (;;)
	{
		/* Echo all received data on the second CDC interface */
		int16_t ReceivedByte = CDC_Device_ReceiveByte(&VirtualSerial2_CDC_Interface);
		if (!(ReceivedByte < 0))
		{
			console_processbyte((uint8_t)ReceivedByte);
			//CDC_Device_SendByte(&VirtualSerial2_CDC_Interface, (uint8_t)ReceivedByte);
		}
		
		//printf("*");

		if (RNDIS_Device_IsPacketReceived(&Ethernet_RNDIS_Interface))
		{
			LEDs_SetAllLEDs(LEDMASK_USB_BUSY);
			
			RNDIS_Device_ReadPacket(&Ethernet_RNDIS_Interface, &FrameIN.FrameData, &FrameIN.FrameLength);
			Ethernet_ProcessPacket(&FrameIN, &FrameOUT);
			
			if (FrameOUT.FrameLength)
			{
				RNDIS_Device_SendPacket(&Ethernet_RNDIS_Interface, &FrameOUT.FrameData, FrameOUT.FrameLength);
				FrameOUT.FrameLength = 0;
			}
			
			LEDs_SetAllLEDs(LEDMASK_USB_READY);
		}

		TCP_TCPTask(&Ethernet_RNDIS_Interface, &FrameOUT);

		RNDIS_Device_USBTask(&Ethernet_RNDIS_Interface);
		CDC_Device_USBTask(&VirtualSerial2_CDC_Interface);
		USB_USBTask();

		if(!(PIND&(1<<0)))
		{
			rfm22b_interrupt();
		}
		while(count_1khz>0)
		{
			count_1khz--;
			tick_1khz();
			count_for_1hz++;
			if(count_for_1hz>=1000)
			{
				count_for_1hz-=1000;
				tick_1hz();
			}
		}
	}
}

void console_sendbyte(char byte)
{
//	if(byte==10) CDC_Device_SendByte(&VirtualSerial2_CDC_Interface, 13);
	CDC_Device_SendByte(&VirtualSerial2_CDC_Interface, byte);
}

/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <joerg@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.        Joerg Wunsch
 * ----------------------------------------------------------------------------
 *
 * Stdio demo
 *
 */

#define RX_BUFSIZE 80

void console_processbyte(char c)
{
//	console_sendbyte(byte+1);
  //uint8_t c;
  char *cp2;
  static char b[RX_BUFSIZE];
  static char *cp = b;
  //static char *rxp;
	//static char *cp3;

	/* behaviour similar to Unix stty ICRNL */
	if (c == '\r')
	  c = '\n';
	if (c == '\n') //enter
	  {
	    console_sendbyte('\r');
	    console_sendbyte('\n');
	    *cp='\0';
            console_processline(b);
	    cp=b;
	    //break;
	  }
	else if (c == '\t') //tab
	  c = ' ';

	if ((c >= (uint8_t)' ' && c <= (uint8_t)'\x7e') ||
	    c >= (uint8_t)'\xa0')
	  {
	    if (cp == b + RX_BUFSIZE - 1)
	      console_sendbyte('\a');
	    else
	      {
		*cp++ = c;
		console_sendbyte(c);
	      }
	    //continue;
	  }

	switch (c)
	  {
	  case 'c' & 0x1f:
	    console_break();
            break;

	  case '\b': //backspace
	  case '\x7f':
	    if (cp > b)
	      {
		console_sendbyte('\b');
		console_sendbyte(' ');
		console_sendbyte('\b');
		cp--;
	      }
	    break;

	  case 'r' & 0x1f: //redraw
	    console_sendbyte('\r');
	    for (cp2 = b; cp2 < cp; cp2++)
	      console_sendbyte(*cp2);
	    break;

	  case 'u' & 0x1f: //erase line
	    while (cp > b)
	      {
		console_sendbyte('\b');
		console_sendbyte(' ');
		console_sendbyte('\b');
		cp--;
	      }
	    break;

	  case 'w' & 0x1f: //erase word
	    while (cp > b && cp[-1] != ' ')
	      {
		console_sendbyte('\b');
		console_sendbyte(' ');
		console_sendbyte('\b');
		cp--;
	      }
	    break;
	  }
}

uint8_t cmdcmp(char *line, char *cmd)
{
	int len;
	len=strlen(cmd);
	if(!strncmp(cmd,line,len))
		return len;
	else
		return 0;
}
uint8_t cmdcmp_P(char *line, PGM_P cmd)
{
	char tempcmd[15];
	uint8_t i;
	i=0;
	while(1)
	{
		tempcmd[i]=pgm_read_byte(cmd+i);
		if(tempcmd[i]==0) break;
		if(i>=15) return 0;
		i++;
	}
	return cmdcmp(line,tempcmd);
}

void console_write(char *word)
{
	char hex[3]="00";
	uint8_t i;
	int b;
	uint8_t a;
	i=0;
	while(word[i]!=' ')
	{
		if(word[i]==0) return;
		i++;
	}
	word[i]=0;
	if(sscanf(word,"%x",&b)<=0)
		return;
	PORTB&=~(1<<0);
	if(noisy>=6) printf_P(PSTR("CS-\r\n"));
	spi_xfer(b | (1<<7));
	word+=i+1;
	while(word[0]!=0)
	{
		hex[0]=word[0];
		if(word[1]==0)
		{
			hex[1]=word[0];
			hex[0]='0';
			word-=1; //hack to make word+2=0;
		}else
		{
			hex[1]=word[1];
		}
		if(sscanf(hex,"%x",&b)>0)
		{
			a=b;
		}else
		{
			a=0;
			if(noisy>=4) printf_P(PSTR("Could not parse \"%s\"\r\n"),hex);
		}
		//printf_P(PSTR("Value: %d\r\n"),a);
		spi_xfer(a);
		word+=2;
	}
	PORTB|=(1<<0);
	if(noisy>=6) printf_P(PSTR("CS+\r\n"));
}
void console_read(char *word)
{
	uint8_t a;
	int b;
	int length=1;
	a=0;
	while(word[a]!=' ')
	{
		if(word[a]==0) break;
		a++;
	}
	if(word[a]==' ' && word[a+1]!=0)
	{
		if(sscanf(word+a+1,"%d",&b)>0)
			length=b;
		word[a]=0;
	}
	if(sscanf(word,"%x",&b)<=0)
		return;
	printf_P(PSTR("Read(%02x): "),b);
	PORTB&=~(1<<0);
	if(noisy>=6) printf_P(PSTR("CS-\r\n"));
	spi_xfer(b);
	while(length>0)
	{
		a=spi_xfer(0);
		printf_P(PSTR("%02x "),a);
		length--;
	}
	PORTB|=(1<<0);
	if(noisy>=6) printf_P(PSTR("CS+\r\n"));
	printf_P(PSTR("\r\n"),a);
}

#define BINARY_COUNT 20

void console_processline(char *line)
{
	uint8_t a;
	int b;
	char *param;
	char *pa;
	char *pb;
	//char hex[3]="00";
	uint8_t i;
	uint8_t c;
	uint8_t binary[BINARY_COUNT];
	float afloat;
	//printf_P(PSTR("Line: \"%s\"\r\n"),line);
	if( (a=cmdcmp_P(line,PSTR("hello"))) )
	{
		param=line+a;
		if(param[0]!=' ')
		{
			printf_P(PSTR("Hello.\r\n"));
		}else
		{
			printf_P(PSTR("Hello: \"%s\"\r\n"),param+1);
		}
	}else if( (a=cmdcmp_P(line,PSTR("nothing"))) )
	{
		param=line+a;
                if(param[0]!=' ')
                {
                        printf_P(PSTR("Usage\r\n"));
                }else
                {
                        param++;
                }
	}else if( (a=cmdcmp_P(line,PSTR("rssi"))) )
	{
		param=line+a;
                if(param[0]!=' ')
                {
			//yeehaw
			afloat=(rssi/((float)256.0));
                        printf_P(PSTR("RSSI: %f (%fdBm)\r\n"), afloat, afloat/2-120);
                }else
                {
                        param++;
			sscanf(param,"%f",&afloat);
			printf_P(PSTR("%f is maybe %fdBm\r\n"),afloat,afloat/2-120);
                }
	}else if( (a=cmdcmp_P(line,PSTR("senda"))) )
	{
		param=line+a;
                if(param[0]!=' ')
                {
                        printf_P(PSTR("Usage\r\n"));
                }else
                {
                        param++;
			a=strlen(param);
			rfm22b_write(0x3e,a); //packet length
			//clear tx fifo
			rfm22b_sbi(8,0);
			rfm22b_cbi(8,0);

			rfm22b_writem(0x7f,(void*)param,a);
			rfm22b_sbi(7,3); //tx TODO not rx at same time
                }
	}else if( (a=cmdcmp_P(line,PSTR("noisy"))) )
	{
		param=line+a;
                if(param[0]==' ' && param[1]!=0)
                {
                        param++;
			sscanf(param,"%d",&b);
			noisy=b;
                }
		printf_P(PSTR("Debug level: %d\r\n"),noisy);
	}else if( (a=cmdcmp_P(line,PSTR("write"))) )
	{
		param=line+a;
                if(param[0]!=' ')
                {
                        printf_P(PSTR("Usage\r\n"));
                }else
                {
                        param++;
			console_write(param);
                }
	}else if( (a=cmdcmp_P(line,PSTR("read"))) )
	{
		param=line+a;
                if(param[0]!=' ')
                {
                        printf_P(PSTR("Usage\r\n"));
                }else
                {
                        param++;
			console_read(param);
                }
	}else if( (a=cmdcmp_P(line,PSTR("rfm22b_init"))) )
	{
		param=line+a;
                if(param[0]!=' ')
                {
			if(rfm22b_init())
			{
				printf_P(PSTR("Fail\r\n"));
			}else
			{
				printf_P(PSTR("Ok\r\n"));
			}

                }else
                {
                        param++;
                }
	}else if( (a=cmdcmp_P(line,PSTR("rfm22b_reset"))) )
	{
		param=line+a;
                if(param[0]!=' ')
                {
			rfm22b_reset();
                }else
                {
                        param++;
                }
	}else if( (a=cmdcmp_P(line,PSTR("parse"))) )
        {
                param=line+a;
                if(param[0]!=' ')
                {
                        printf_P(PSTR("Usage\r\n"));
                }else
                {
                        param++;
			pa=strtok_r(param," ",&pb);
			c=0;
			while(pa != NULL && c<BINARY_COUNT)
			{
				c+=parse_ascii(pa,binary+c,BINARY_COUNT-c);
				pa=strtok_r(NULL," ",&pb);
			}
			printf_P(PSTR("Result (%d): "),c);
			for(i=0;i<c;i++)
				printf_P(PSTR("%d "),binary[i]);
			printf_P(PSTR("\r\n"),c);
                }
	}else if( (a=cmdcmp_P(line,PSTR("regdump"))) )
	{
		param=line+a;
                if(param[0]!=' ')
                {
                        for(i=0;i<=0x7E;i++)
			{
				//this code is dumb because all the registers can be read at once
				PORTB&=~(1<<0);
				if(noisy>=6) printf_P(PSTR("CS-\r\n"));
				SPDR=i;
				while(!(SPSR & (1<<SPIF)));
				SPDR=0;
				while(!(SPSR & (1<<SPIF)));
				printf_P(PSTR("Add %02x Value %02x\r\n"),i,SPDR);
				PORTB|=(1<<0);
				if(noisy>=6) printf_P(PSTR("CS+\r\n"));
			}
                }else
                {
                        param++;
                }
	}else if( (a=cmdcmp_P(line,PSTR("spi"))) )
        {
                param=line+a;
                if(param[0]!=' ')
                {
                        printf_P(PSTR("Usage\r\n"));
                }else
                {
                        param++;
			pa=strtok_r(param," ",&pb);
			c=0;
			while(pa != NULL && c<BINARY_COUNT)
			{
				c+=parse_ascii(pa,binary+c,BINARY_COUNT-c);
				pa=strtok_r(NULL," ",&pb);
			}
			printf_P(PSTR("Sending (%d): "),c);
			PORTB&=~(1<<0);
			if(noisy>=6) printf_P(PSTR("CS-\r\n"));
			for(i=0;i<c;i++)
			{
				SPDR=binary[i];
				printf_P(PSTR("%x-"),binary[i]);
				while(!(SPSR & (1<<SPIF)));
				printf_P(PSTR("%x "),SPDR);
			}
			PORTB|=(1<<0);
			if(noisy>=6) printf_P(PSTR("CS+\r\n"));
			printf_P(PSTR("\r\n"),c);
                }
        }

}

uint8_t parse_ascii(char *word, uint8_t *store, uint8_t max)
{
	uint8_t count=0;
	int b;
	uint8_t a;
	uint8_t c;
	uint8_t i;
	char hex[3]="00";

	//printf_P(PSTR("parse_ascii(%s,%d,%d)\r\n"),word,store,max);

	if(word[0]=='0' && word[1]=='x' && word[2]!=0)
	{
		word+=2;
		while(word[0]!=0)
		{
			hex[0]=word[0];
			if(word[1]==0)
			{
				hex[1]=word[0];
				hex[0]='0';
				word-=1; //hack to make word+2=0;
			}else
			{
				hex[1]=word[1];
			}
			if(sscanf(hex,"%x",&b)>0)
			{
				a=b;
			}else
			{
				a=0;
				if(noisy>=4) printf_P(PSTR("Could not parse \"%s\"\r\n"),hex);
			}
			//printf_P(PSTR("Value: %d\r\n"),a);
			store[count++]=a;
			if(count>=max) return count;
			word+=2;
		}
	}else if(word[0]=='0' && word[1]=='b' && word[2]!=0)
	{
		word+=2;
		c=0;
		while(1)
		{
			a=0;
			for(i=0;i<8;i++)
			{
				//printf_P(PSTR("B(%d,%x) "),i,word[i]);
				if(word[i]==0)
				{
					c=1;
				 	break;
				}
				if(word[i]!='0') a|=1<<(7-i);
				//a |= (word[i]?1:0) << (7-i);
			}
			if(c==1) break;
			word+=8;
			//printf_P(PSTR("Value: %d\r\n"),a);
			store[count++]=a;
			if(count>=max) return count;
		}
	}else if(word[0]=='0' && word[1]=='a' && word[2]!=0)
	{
		word+=2;
		while(*word!=0)
		{
			//printf_P(PSTR("Value: %d\r\n"),*word);
			store[count++]=*word;
			if(count>=max) return count;
			word++;
		}
	}else
	{
		if(sscanf(word,"%d",&b)>0)
		{
			a=b;
		}else
		{
			a=0;
			if(noisy>=4) printf_P(PSTR("Could not parse \"%s\"\r\n"),word);
			}
		//printf_P(PSTR("Value: %d\r\n"),a);
		store[count++]=a;
		if(count>=max) return count;
	}
	return count;
}

/** Configures the board hardware and chip peripherals for the demo's functionality. */
void SetupHardware(void)
{
	/* Disable watchdog if enabled by bootloader/fuses */
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	/* Disable clock division */
	clock_prescale_set(clock_div_2);

	DDRA=0;
	PORTA=0xFF;
	DDRB=(1<<0)|(1<<1)|(1<<2)|(1<<4); //spi, reset
	PORTB=(1<<0)|(1<<4)|(1<<5)|(1<<6)|(1<<7);
	DDRC=0;
	PORTC=0xFF;
	DDRD=0;
	PORTD=0b01111111;
	DDRE=0;
	PORTE=0xFF;
	DDRF=0;
	PORTF=0xFF;

	SPCR=(1<<SPE)|(1<<MSTR);
	//SPCR=(1<<SPE)|(1<<MSTR)|(1<<SPR1)|(1<<SPR0);
	SPSR=(1<<SPI2X);

	OCR0A=124; // /125
	TCCR0A=(1<<WGM01); //ctc for OCR0A
	TCCR0B=(1<<CS01)|(1<<CS00); // /64
	TIMSK0=(1<<OCIE0A); //compare match interrupt

	/* Hardware Initialization */
	LEDs_Init();
	//Serial_Init(9600, false);
	USB_Init();

	/* Create a stdio stream for the serial port for stdin and stdout */
	//Serial_CreateStream(NULL);
	CDC_Device_CreateStream(&VirtualSerial2_CDC_Interface, &USBSerialStream);
	stdin=stdout=stderr=&USBSerialStream;

	printf_P(PSTR("Hello, World!\r\n"));

	//rfm22b_reset();
	//if(rfm22b_init())
	//{
	//	printf_P(PSTR("rfm22b init failure\r\n"));
	//}
}

/** Event handler for the library USB Connection event. */
void EVENT_USB_Device_Connect(void)
{
	LEDs_SetAllLEDs(LEDMASK_USB_ENUMERATING);
}

/** Event handler for the library USB Disconnection event. */
void EVENT_USB_Device_Disconnect(void)
{
	LEDs_SetAllLEDs(LEDMASK_USB_NOTREADY);
}

/** Event handler for the library USB Configuration Changed event. */
void EVENT_USB_Device_ConfigurationChanged(void)
{
	bool ConfigSuccess = true;

	ConfigSuccess &= RNDIS_Device_ConfigureEndpoints(&Ethernet_RNDIS_Interface);
	ConfigSuccess &= CDC_Device_ConfigureEndpoints(&VirtualSerial2_CDC_Interface);

	LEDs_SetAllLEDs(ConfigSuccess ? LEDMASK_USB_READY : LEDMASK_USB_ERROR);
}

/** Event handler for the library USB Control Request reception event. */
void EVENT_USB_Device_ControlRequest(void)
{
	RNDIS_Device_ProcessControlRequest(&Ethernet_RNDIS_Interface);
	CDC_Device_ProcessControlRequest(&VirtualSerial2_CDC_Interface);
}

void rfm22b_reset()
{
	PORTB|=(1<<4); //hardware reset
	_delay_ms(16);
	PORTB&=~(1<<4);
	_delay_ms(16);
	rfm22b_write(0x07,0x80); //software reset
	_delay_ms(16);
}

int8_t rfm22b_init()
{
	uint8_t i;
	uint8_t addr;
	uint8_t val;
	uint8_t a;

	i=0;
	if(rfm22b_read(0)!=0x08 || rfm22b_read(1)!=0x06)
	{
		if(noisy>=2) printf_P(PSTR("communication error\r\n"));
		return 3;
	}
	/*i=0;
	while(!(rfm22b_read(0x04)&(1<<1)))
	{
		i++;
		if(noisy>=2) printf_P(PSTR("chip not ready\r\n"));
		_delay_ms(100);
		if(i>5) return 2;
	}*/
	//_delay_ms(16); //todo
	for(i=0;i<RFM22B_REGS_COUNT;i++)
	{
		addr=pgm_read_byte(&rfm22b_regs[i*2]);
		val=pgm_read_byte(&rfm22b_regs[i*2+1]);
		if(noisy>=6) printf_P(PSTR("num %d addr %02x val %02x\r\n"),i,addr,val);
		rfm22b_writem(addr,&val,1);
		rfm22b_readm(addr,&a,1);
		if(a!=val)
		{
			printf_P(PSTR("error writing register\r\n"));
			return 1;
		}
	}
	return 0;
}

void rfm22b_writem(uint8_t reg, uint8_t *data, uint8_t count)
{
	uint8_t i;
	if(noisy>=6) printf_P(PSTR("rfm22b_writem(%02x,%d,%d)\r\n"),reg,data,count);
	PORTB&=~(1<<0);
	if(noisy>=6) printf_P(PSTR("CS-\r\n"));
	spi_xfer(reg | (1<<7));
	for(i=0;i<count;i++)
		spi_xfer(data[i]);
	PORTB|=(1<<0);
	if(noisy>=6) printf_P(PSTR("CS+\r\n"));
}

void rfm22b_readm(uint8_t reg, uint8_t *data, uint8_t count)
{
	uint8_t i;
	if(noisy>=6) printf_P(PSTR("rfm22b_readm(%02x,%d,%d)\r\n"),reg,data,count);
        PORTB&=~(1<<0);
	if(noisy>=6) printf_P(PSTR("CS-\r\n"));
        spi_xfer(reg);
        for(i=0;i<count;i++)
                data[i]=spi_xfer(0);
        PORTB|=(1<<0);
	if(noisy>=6) printf_P(PSTR("CS+\r\n"));
}

uint8_t rfm22b_read(uint8_t reg)
{
	uint8_t a;
	if(noisy>=6) printf_P(PSTR("rfm22b_read(%02x)\r\n"),reg);
        PORTB&=~(1<<0);
	if(noisy>=6) printf_P(PSTR("CS-\r\n"));
        spi_xfer(reg);
	a=spi_xfer(0);
        PORTB|=(1<<0);
	if(noisy>=6) printf_P(PSTR("CS+\r\n"));
	return a;
}

void rfm22b_write(uint8_t reg, uint8_t data)
{
	if(noisy>=6) printf_P(PSTR("rfm22b_write(%02x,%02x)\r\n"),reg,data);
	PORTB&=~(1<<0);
	if(noisy>=6) printf_P(PSTR("CS-\r\n"));
	spi_xfer(reg | (1<<7));
	spi_xfer(data);
	PORTB|=(1<<0);
	if(noisy>=6) printf_P(PSTR("CS+\r\n"));
}

uint8_t spi_xfer(uint8_t send)
{
	if(noisy>=7) printf_P(PSTR("spi_xfer(%02x)..."),send);
	//while(!(SPSR & (1<<SPIF)));
	SPDR=send;
	while(!(SPSR & (1<<SPIF)));
	if(noisy>=7) printf_P(PSTR("%02x\r\n"),SPDR);
	return SPDR;
}

void console_break()
{
	printf_P(PSTR("Break\r\n"));
}

void rfm22b_interrupt()
{
	uint8_t status,flags1,flags2;
	uint8_t a;
	status=rfm22b_read(2);
	flags1=rfm22b_read(3);
	flags2=rfm22b_read(4);
	rf_packet packet;
	uint8_t data_length=0;
	if(noisy>=5) printf_P(PSTR("Interrupt status: %02x %02x %02x\r\n"),status,flags1,flags2);
	if(flags2&(1<<0))
	{
		if(noisy>=3) printf_P(PSTR("Power on, running init\r\n"));
		rfm22b_init();
		return;
	}
	if(status&0xC0 || flags1&(1<<7))
	{
		if(noisy>=1) printf_P(PSTR("FIFO%s%s error, resetting fifos\r\n"),(status&(1<<7))?" overflow":"",(status&(1<<6))?" underflow":"");
		a=rfm22b_read(8);
		a|=0x03;
		rfm22b_write(8,a);
		a&=~(0x03);
		rfm22b_write(8,a);
	}
	if(flags1&(1<<2))
	{
		if(noisy>=5) printf_P(PSTR("Packet Sent\r\n"));
		rfm_status.transmitter_active=0;
	}
	if(flags1&(1<<1))
	{
		if(noisy>=5) printf_P(PSTR("Packet Received\r\n"));
		rfm22b_cbi(7,2);
		while( !(rfm22b_read(2)&(1<<5)) )
		{
			rfm22b_readm(0x7f,(void*)&packet,5);
			if(packet.length>data_length)
			{
				if(packet.length>MAX_PACKET_LENGTH)
				{
					if(noisy>=2) printf_P(PSTR("Length invalid\r\n"));
					//clear rx because we won't be reading it
					rfm22b_sbi(8,1);
					rfm22b_cbi(8,1);

				}else
				{
					packet.data=alloca(packet.length); //two packets recieved is bad
					data_length=packet.length;
				}
			}
			if(packet.length<=data_length) //don't continue on failure from above
			{
				rfm22b_readm(0x7f,packet.data,packet.length);
				printf_P(PSTR("Header: To %02x From %02x type %02x seq %02x Len %d\r\n"), packet.destination, packet.source, packet._type, packet.sequence, packet.length);
				if(packet.length>0) printf_P(PSTR("Data: "));
				for(a=0;a<packet.length;a++)
				{
					printf_P(PSTR("%02x"),packet.data[a]);
				}
				if(packet.length>0) printf_P(PSTR("\r\n"));
			}
		}
		//todo enable rx or idle state chooser
	}
	if(flags1&(1<<0))
	{
		if(noisy>=3) printf_P(PSTR("CRC Error\r\n"));
	}
	if(flags2&(1<<7))
	{
		a=rfm22b_read(0x26);
		if(noisy>=5) printf_P(PSTR("Sync RSSI: %d\r\n"),a);
	}
}
inline void rfm22b_cbi(uint8_t reg, uint8_t bv)
{
	rfm22b_cb(reg,(1<<bv));
}
inline void rfm22b_sbi(uint8_t reg, uint8_t bv)
{
	rfm22b_sb(reg,(1<<bv));
}
void rfm22b_cb(uint8_t reg, uint8_t mask)
{
	uint8_t a;
	a=rfm22b_read(reg);
	a&=~mask;
	rfm22b_write(reg,a);
}
void rfm22b_sb(uint8_t reg, uint8_t mask)
{
	uint8_t a;
	a=rfm22b_read(reg);
	a|=mask;
	rfm22b_write(reg,a);
}

ISR(TIMER0_COMPA_vect)
{
	count_1khz++;
}

void tick_1khz(void)
{
	//TODO don't randomly sample rssi for noise level indication
	uint8_t a;
	a=rfm22b_read(0x26);
	rssi_decimate+=a;
	rssi_decimate_count++;
	if(rssi_decimate_count==255)
	{
		rssi=rssi_decimate;
		rssi_decimate=0;
	}
}
void tick_1hz(void)
{
	if(noisy>=7) printf_P(PSTR("Tick\r\n"));
}

